# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class ExecuteCommandResponse(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'command_id': 'str',
        'directory_name': 'str',
        'dns_address': 'str'
    }

    attribute_map = {
        'command_id': 'command_id',
        'directory_name': 'directory_name',
        'dns_address': 'dns_address'
    }

    def __init__(self, command_id=None, directory_name=None, dns_address=None, local_vars_configuration=None):  # noqa: E501
        """ExecuteCommandResponse - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._command_id = None
        self._directory_name = None
        self._dns_address = None
        self.discriminator = None

        self.command_id = command_id
        self.directory_name = directory_name
        self.dns_address = dns_address

    @property
    def command_id(self):
        """Gets the command_id of this ExecuteCommandResponse.  # noqa: E501


        :return: The command_id of this ExecuteCommandResponse.  # noqa: E501
        :rtype: str
        """
        return self._command_id

    @command_id.setter
    def command_id(self, command_id):
        """Sets the command_id of this ExecuteCommandResponse.


        :param command_id: The command_id of this ExecuteCommandResponse.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and command_id is None:  # noqa: E501
            raise ValueError("Invalid value for `command_id`, must not be `None`")  # noqa: E501

        self._command_id = command_id

    @property
    def directory_name(self):
        """Gets the directory_name of this ExecuteCommandResponse.  # noqa: E501


        :return: The directory_name of this ExecuteCommandResponse.  # noqa: E501
        :rtype: str
        """
        return self._directory_name

    @directory_name.setter
    def directory_name(self, directory_name):
        """Sets the directory_name of this ExecuteCommandResponse.


        :param directory_name: The directory_name of this ExecuteCommandResponse.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and directory_name is None:  # noqa: E501
            raise ValueError("Invalid value for `directory_name`, must not be `None`")  # noqa: E501

        self._directory_name = directory_name

    @property
    def dns_address(self):
        """Gets the dns_address of this ExecuteCommandResponse.  # noqa: E501


        :return: The dns_address of this ExecuteCommandResponse.  # noqa: E501
        :rtype: str
        """
        return self._dns_address

    @dns_address.setter
    def dns_address(self, dns_address):
        """Sets the dns_address of this ExecuteCommandResponse.


        :param dns_address: The dns_address of this ExecuteCommandResponse.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and dns_address is None:  # noqa: E501
            raise ValueError("Invalid value for `dns_address`, must not be `None`")  # noqa: E501

        self._dns_address = dns_address

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ExecuteCommandResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ExecuteCommandResponse):
            return True

        return self.to_dict() != other.to_dict()



from contextlib import contextmanager
import logging
import os
import re
import subprocess
import sys
import tempfile
from threading import Event, Thread, Timer
import time
from typing import Dict, Iterator, List, Optional

from openapi_client.rest import ApiException  # type: ignore
import ray

import anyscale
from anyscale.client.openapi_client.api.default_api import DefaultApi  # type: ignore
from anyscale.util import format_api_exception, get_rsync_command, send_json_request

logging.basicConfig(format=ray.ray_constants.LOGGER_FORMAT)
logger = logging.getLogger(__file__)
logging.getLogger("botocore").setLevel(logging.CRITICAL)

if anyscale.conf.AWS_PROFILE is not None:
    logger.info("Using AWS profile %s", anyscale.conf.AWS_PROFILE)
    os.environ["AWS_PROFILE"] = anyscale.conf.AWS_PROFILE

TIME_BETWEEN_HEARTBEATS = 30  # seconds


@contextmanager
def managed_autosync_session(
    session_id: str, api_client: DefaultApi = None
) -> Iterator[str]:
    # Register to the API that we enabled autosync
    with format_api_exception(ApiException):
        resp = api_client.register_autosync_session_api_v2_autosync_sessions_post(
            session_id
        )

    autosync_session_id = resp.result.id
    heartbeat_thread = AutosyncHeartbeat(autosync_session_id)
    heartbeat_thread.start()

    try:
        yield autosync_session_id
    finally:
        heartbeat_thread.finish.set()
        with format_api_exception(ApiException):
            api_client.deregister_autosync_session_api_v2_autosync_sessions_autosync_session_id_delete(
                autosync_session_id
            )
        heartbeat_thread.join()
        print("Autosync finished.")


class AutosyncHeartbeat(Thread):
    def __init__(self, autosync_session_id: str):
        super().__init__()
        self.autosync_session_id = autosync_session_id
        self.finish = Event()

    def run(self) -> None:
        while not self.finish.is_set():
            try:
                heartbeat_status_str = "Last heartbeat sent at {}".format(
                    time.strftime("%x %X %Z")
                )  # Locale date, locale time, timezone string.

                # Keep the heartbeat status output a single line.
                heartbeat_status_str = "\033[1A\033[0K" + heartbeat_status_str
                print(heartbeat_status_str)
                send_json_request(
                    "/api/v2/autosync_sessions/{}/heartbeat".format(
                        self.autosync_session_id
                    ),
                    {},
                    method="POST",
                )
            except Exception as e:
                print("Error sending heartbeat:", e)
            self.finish.wait(TIME_BETWEEN_HEARTBEATS)


def _do_fullsync_autosync(
    ssh_command: List[str],
    source: str,
    ssh_user: str,
    head_ip: str,
    target: str,
    sync_git: bool,
) -> None:
    anyscale_dir = os.path.dirname(os.path.realpath(__file__))
    if sys.platform.startswith("linux"):
        env: Dict[str, str] = {}
        fswatch_executable = os.path.join(anyscale_dir, "fswatch-linux")
        fswatch_command = [
            fswatch_executable,
            source,
            "--monitor=poll_monitor",
            "-o",
        ]
    elif sys.platform.startswith("darwin"):
        env = {"DYLD_LIBRARY_PATH": anyscale_dir}
        fswatch_executable = os.path.join(anyscale_dir, "fswatch-darwin")
        fswatch_command = [
            fswatch_executable,
            source,
            "-o",
        ]
    else:
        raise NotImplementedError(
            "Autosync not supported on platform {}".format(sys.platform)
        )

    # Perform synchronization whenever there is a change. We batch together
    # multiple updates and then call rsync on them.
    with subprocess.Popen(
        fswatch_command, stdout=subprocess.PIPE, env=env,
    ) as proc:  # noqa: B1

        def do_sync() -> None:
            print("syncing...")
            command, rsync_env = get_rsync_command(
                ssh_command,
                source,
                ssh_user,
                head_ip,
                target,
                sync_git,
                modified_files=None,
            )
            logger.info("Calling rsync due to detected file update.")
            logger.debug("Command: {command}".format(command=command))
            temp_sync = subprocess.run(command, env=rsync_env)  # noqa: B1
            if temp_sync.returncode != 0:
                logger.error(
                    f"rsync failed with error code: {temp_sync.returncode}, autosync will continue running."
                )

        while True:
            timer: Optional[Timer] = None
            while proc.stdout:
                proc.stdout.readline()
                if timer is not None:
                    timer.cancel()

                timer = Timer(2, do_sync)
                timer.start()


def _do_iterative_sync_autosync(
    ssh_command: List[str],
    source: str,
    ssh_user: str,
    head_ip: str,
    target: str,
    sync_git: bool,
) -> None:
    anyscale_dir = os.path.dirname(os.path.realpath(__file__))
    if sys.platform.startswith("linux"):
        env: Dict[str, str] = {}
        fswatch_executable = os.path.join(anyscale_dir, "fswatch-linux")
        fswatch_command = [
            fswatch_executable,
            source,
            "--batch-marker",
            "--monitor=poll_monitor",
            "-r",
        ]
        if not sync_git:
            fswatch_command += ["-e", ".*/\.git"]  # noqa: W605
    elif sys.platform.startswith("darwin"):
        env = {"DYLD_LIBRARY_PATH": anyscale_dir}
        fswatch_executable = os.path.join(anyscale_dir, "fswatch-darwin")
        fswatch_command = [
            fswatch_executable,
            source,
            "--batch-marker",
        ]
        if not sync_git:
            fswatch_command += ["-e", ".*/\.git"]  # noqa: W605
    else:
        raise NotImplementedError(
            "Autosync not supported on platform {}".format(sys.platform)
        )

    # Check if current directory is a git repo:
    try:
        subprocess.check_call(  # noqa: B1
            ["git", "-C", ".", "rev-parse"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        is_git_repo = True
    except subprocess.CalledProcessError:
        is_git_repo = False

    # Perform synchronization whenever there is a change. We batch together
    # multiple updates and then call rsync on them.
    with subprocess.Popen(
        fswatch_command, stdout=subprocess.PIPE, env=env,
    ) as proc:  # noqa: B1
        while True:
            files = []
            while proc.stdout:
                path = proc.stdout.readline().strip().decode()
                if path == "NoOp":
                    break
                else:
                    relpath = os.path.relpath(path, source)
                    files.append(relpath)
            if files:
                if is_git_repo:
                    # Filter out gitignore files
                    try:
                        ignored_files = subprocess.check_output(  # noqa: B1
                            ["git", "check-ignore", *files]
                        )
                        files_to_ignore = ignored_files.decode().splitlines()
                        for file in files_to_ignore:
                            files.remove(file)
                    except subprocess.CalledProcessError:
                        # Failed check-ignore just means we won't ignore any files
                        # For example, if the directory is not a git repo, this call will fail.
                        pass

                with tempfile.NamedTemporaryFile(mode="w") as modified_files:
                    for f in files:
                        # Avoid rsyncing temporary files that may disappear between
                        # now and when rsync builds a file list.
                        # This matches files ending with: .swp, ~, .tmp or 4913 (neovim)
                        if re.match("(.*(.sw[a-e,g-z]|~|.tmp)$|^4913$)", f):
                            continue
                        modified_files.write(f + "\n")
                    modified_files.flush()
                    command, rsync_env = get_rsync_command(
                        ssh_command,
                        source,
                        ssh_user,
                        head_ip,
                        target,
                        sync_git,
                        modified_files=modified_files,
                    )
                    logger.info("Calling rsync due to detected file update.")
                    logger.debug("Command: {command}".format(command=command))
                    temp_sync = subprocess.run(command, env=rsync_env)  # noqa: B1
                    if temp_sync.returncode != 0:
                        logger.error(
                            f"Secondary sync failed with error code: {temp_sync.returncode}, autosync will continue running."
                        )


def perform_autosync_synchronization(
    ssh_command: List[str],
    source: str,
    ssh_user: str,
    head_ip: str,
    target: str,
    sync_git: bool,
    full_sync: bool,
) -> None:
    if full_sync:
        _do_fullsync_autosync(ssh_command, source, ssh_user, head_ip, target, sync_git)
    else:
        _do_iterative_sync_autosync(
            ssh_command, source, ssh_user, head_ip, target, sync_git
        )



import json
import logging
import os
import sys
from typing import Any

import click
from packaging import version
import ray.ray_constants

from anyscale.api import instantiate_api_client
from anyscale.client.openapi_client.rest import ApiException  # type: ignore
from anyscale.cloudgateway import CloudGatewayRunner
from anyscale.commands.cloud_commands import cloud_cli
from anyscale.commands.exec_commands import anyscale_exec
from anyscale.commands.list_commands import list_cli
from anyscale.commands.project_commands import anyscale_clone, anyscale_init
from anyscale.commands.session_commands import (
    anyscale_autosync,
    anyscale_fork,
    anyscale_pull,
    anyscale_push,
    anyscale_ssh,
    anyscale_stop,
    anyscale_up,
)
from anyscale.commands.session_commands_hidden import (
    anyscale_start,
    session_cli,
)
from anyscale.commands.snapshot_commands_hidden import snapshot_cli
import anyscale.conf
from anyscale.util import send_json_request


logging.basicConfig(format=ray.ray_constants.LOGGER_FORMAT)
logger = logging.getLogger(__file__)
logging.getLogger("botocore").setLevel(logging.CRITICAL)

if anyscale.conf.AWS_PROFILE is not None:
    logger.info("Using AWS profile %s", anyscale.conf.AWS_PROFILE)
    os.environ["AWS_PROFILE"] = anyscale.conf.AWS_PROFILE


class AliasedGroup(click.Group):
    # This is from https://stackoverflow.com/questions/46641928/python-click-multiple-command-names
    def get_command(self, ctx: Any, cmd_name: str) -> Any:
        try:
            cmd_name = ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


@click.group(
    invoke_without_command=True,
    no_args_is_help=True,
    cls=AliasedGroup,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option(
    "--version",
    "-v",
    "version_flag",
    is_flag=True,
    default=False,
    help="Current anyscale version.",
)
@click.option(
    "--json",
    "show_json",
    is_flag=True,
    default=False,
    help="Return output as json, for use with --version.",
)
@click.pass_context
def cli(ctx: Any, version_flag: bool, show_json: bool) -> None:
    try:
        api_client = instantiate_api_client(no_cli_token=True)
        resp = api_client.get_anyscale_version_api_v2_userinfo_anyscale_version_get()
        curr_version = anyscale.__version__
        latest_version = resp.result.version
        if version.parse(curr_version) < version.parse(latest_version):
            message = "Warning: Using version {0} of anyscale. Please update the package using pip install anyscale -U to get the latest version {1}".format(
                curr_version, latest_version
            )
            print("\033[91m{}\033[00m".format(message), file=sys.stderr)
    except ApiException as e:
        logger.warning(
            "Error {} while trying to get latest anyscale version number: {}".format(
                e.status, e.reason
            )
        )
    except Exception as e:
        logger.warning(e)

    if version_flag:
        ctx.invoke(version_cli, show_json=show_json)


@click.command(name="version", help="Display version of the anyscale CLI.")
@click.option(
    "--json", "show_json", is_flag=True, default=False, help="Return output as json."
)
def version_cli(show_json: bool) -> None:
    if show_json:
        print(json.dumps({"version": anyscale.__version__}))
    else:
        print(anyscale.__version__)


@cli.command(
    name="help", help="Display help documentation for anyscale CLI.", hidden=True
)
@click.pass_context
def anyscale_help(ctx: Any) -> None:
    print(ctx.parent.get_help())


@click.command(
    name="cloudgateway",
    help="Run private clusters via anyscale cloud gateway.",
    hidden=True,
)
@click.option("--gateway-id", type=str, required=True)
def anyscale_cloudgateway(gateway_id: str) -> None:
    # Make sure only registered users can start the gateway.
    logger.info("Verifying user ...")
    try:
        send_json_request("/api/v2/userinfo/", {})
    except Exception:
        raise click.ClickException(
            "Invalid user. Did you set up the cli_token credentials?"
            + ' To setup your credentials, follow the instructions in the "credentials" tab'
            + " after logging in to your anyscale account."
        )
    anyscale_address = f"/api/v2/cloudgateway/{gateway_id}"
    cloudgateway_runner = CloudGatewayRunner(anyscale_address)
    logger.info(
        "Your gateway-id is: {}. Store it in the provider section in the".format(
            gateway_id
        )
        + " cluster yaml file of the remote cluster that interacts with this gateway."
        + ' E.g., config["provider"]["gateway_id"]={gateway_id}.'.format(
            gateway_id=gateway_id
        )
    )
    cloudgateway_runner.gateway_run_forever()


cli.add_command(session_cli)
cli.add_command(snapshot_cli)
cli.add_command(cloud_cli)
cli.add_command(version_cli)
cli.add_command(list_cli)

cli.add_command(anyscale_init)
cli.add_command(anyscale_start)
cli.add_command(anyscale_up)
cli.add_command(anyscale_stop)
cli.add_command(anyscale_cloudgateway)
cli.add_command(anyscale_autosync)
cli.add_command(anyscale_fork)
cli.add_command(anyscale_clone)
cli.add_command(anyscale_ssh)
cli.add_command(anyscale_exec)
cli.add_command(anyscale_pull)
cli.add_command(anyscale_push)
cli.add_command(anyscale_help)

ALIASES = {"h": anyscale_help}


def main() -> Any:
    return cli()


if __name__ == "__main__":
    main()
